from pytest import approx
from energy_tools.per_unit import *


def test_i_base_ka():
    assert i_base_ka(100, 120) == approx(0.481125)


def test_z_base_ohm_ib():
    assert z_base_ohm(120, ib_ka=0.481125) == approx(144)


def test_z_base_ohm_sb():
    assert z_base_ohm(120, sb_mva=100) == approx(144)


def test_s_base_mva():
    assert s_base_mva(120, 0.481125) == approx(100)



from energy_tools.complex import complex
from energy_tools.phasor import Phasor


# TODO(miek770): Add tests for sequences() and phasors().


def test_phasor_addition():
    v1 = Phasor(7.0, 0.0)
    v2 = Phasor(4.0, 45.0)
    v3 = Phasor(10.23, 16.05)
    ans = v1 + v2
    assert ans == v3


def test_phasor_addition_triple():
    v1 = Phasor(10, 20)
    v2 = Phasor(15, 90)
    v3 = Phasor(7, 190)
    v4 = Phasor(17.39, 81.72)
    assert sum((v1, v2, v3)) == v4


def test_phasor_soustraction():
    v1 = Phasor(1.5, 90)
    v2 = Phasor(2.6, 145)
    v3 = Phasor(2.13, 0.24)
    ans = v1 - v2
    print(ans)
    print(v3)
    assert ans == v3


def test_phasor_soustraction_et_addition():
    v1 = Phasor(22, 140)
    v2 = Phasor(40, 190)
    v3 = Phasor(15, 290)
    v4 = Phasor(28.54, 14.18)
    assert v1 - v2 + v3 == v4


def test_phasor_egalite():
    assert Phasor(7.0, 0.0) == Phasor(7.0, 0.0)


def test_phasor_multiplication():
    v1 = Phasor(6, 30)
    v2 = Phasor(8, -45)
    v3 = Phasor(48, -15)
    assert v1 * v2 == v3


def test_phasor_division():
    v1 = Phasor(6, 30)
    v2 = Phasor(8, -45)
    v3 = Phasor(0.75, 75)
    assert v1 / v2 == v3


def test_phasor_puissance_cube():
    v1 = Phasor(2, 10)
    v2 = Phasor(8, 30)
    assert pow(v1, 3) == v2


def test_phasor_puissance_carre():
    v1 = Phasor(20, 270)
    v2 = Phasor(400, 180)
    assert pow(v1, 2) == v2


def test_phasor_puissance_carre_2():
    v1 = Phasor(20, 270)
    v2 = Phasor(400, 180)
    assert v1 ** 2 == v2


def test_phasor_representation_decimal():
    v1 = Phasor(2.14, 127.12)
    assert v1.__repr__() == "2.14 @ 127.12°"


def test_phasor_representation_entier():
    v1 = Phasor(2, 127)
    assert v1.__repr__() == "2.0 @ 127.0°"


def test_phasor_complexe():
    v1 = Phasor(3.0 + 4.0j)
    assert v1.__repr__() == "5.0 @ 53.13°"


def test_phasor_complexe_real():
    v1 = Phasor(3.0 + 4.0j)
    assert v1.real == 3.0


def test_phasor_complexe_imag():
    v1 = Phasor(3.0 + 4.0j)
    assert v1.imag == 4.0


def test_phasor_multiplication_complexe():
    v1 = Phasor(3.0 + 4.0j)
    assert complex(3.0 + 4.0j) * v1 == pow(v1, 2)



from math import atan, pi


class complex(complex):
    """Improvement of the builtin complex type.

    Additional attributes:
        phase: Returns the complex number's phase in degrees.
        module: Returns the complex number's unitless amplitude.
    """

    @property
    def phase(self):
        """Returns the complex number's phase in degrees.
        """
        if self.real > 0:
            if self.imag >= 0:
                return abs(180 * atan(self.imag / self.real) / pi)
            else:
                return 360 - abs(180 * atan(self.imag / self.real) / pi)
        elif self.real < 0:
            if self.imag >= 0:
                return 180 - abs(180 * atan(self.imag / self.real) / pi)
            else:
                return 180 + abs(180 * atan(self.imag / self.real) / pi)
        else:
            if self.imag == 0:
                return 0.0
            elif self.imag > 0:
                return 90.0
            else:
                return 270.0

    @property
    def module(self):
        return abs(self)


def complex_impedance(z, xr):
    """Returns a complex impedance based on an impedance *z* and the X/R ratio.

    Args:
        z: Unitless impedance value.
        xr: X/R ratio.

    Returns:
        Complex impedance.
    """

    z = float(abs(z))
    xr = float(abs(xr))
    real = (z ** 2 / (1 + xr ** 2)) ** 0.5
    try:
        imag = (z ** 2 / (1 + 1 / xr ** 2)) ** 0.5
    except ZeroDivisionError:
        imag = 0.0
    return complex(real, imag)



from math import pi
from energy_tools.complex import complex


def serie(z):
    """Puts impedances in serie (yep, simple summation).

    Args:
        z: List of impedances (either complex or real or a combination of both).

    Returns:
        Serie impedance.
    """
    return sum(z)


def parallel(z):
    """Puts impedances in parallel.

    Args:
        z: List of impedances (either complex or real or a combination of both).

    Returns:
        Parallel impedance.
    """
    try:
        return 1 / sum([1.0 / y for y in z])

    except ZeroDivisionError:
        return 0.0


def zCap(C, f=60):
    """Retourne l'impédance complexe d'un banc de condensateurs en fonction de
    la capacitance et de la fréquence (à 60Hz par défaut).
    """
    return complex(1 / (2j * pi * f * C))


def zInd(L, f=60):
    """Retourne l'impédance complexe d'une inductance en fonction de
    l'inductance et de la fréquence (à 60Hz par défaut).
    """
    return complex(2j * pi * f * L)



from pytest import approx, raises
from energy_tools.misc import parallel, serie


def test_serie_resistances():
    assert serie((2, 3.5)) == 5.5


def test_serie_impedances():
    assert approx(serie((2 + 1j, 3.0 - 2.8j))) == 5 - 1.8j


def test_serie_erreur_type():
    with raises(TypeError):
        serie("test")


def test_parallel_resistances():
    assert parallel((2.4, 1.2)) == 0.8


def test_parallel_impedances():
    ans = parallel((310.2 + 12.4j, 18.3 - 132.7j))
    assert ans.real + ans.imag * 1j == approx(59.893270 - 102.683286j)



def i_base_ka(sb_mva, vbll_kv):
    """Returns the base current based on the base power and voltage.

    Args:
        sb_mva: Base power in MVA.
        vbll_kv: Base voltage in kV.

    Returns:
        Base current in kA.
    """
    sb_mva = float(sb_mva)
    vbll_kv = float(vbll_kv)
    return sb_mva / (vbll_kv * 3 ** 0.5)


def z_base_ohm(vbll_kv, ib_ka=None, sb_mva=None):
    """Returns the base impedance based on the base voltage and either the base current
    or the base power.

    Args:
        vbll_kv: Base voltage in kV.
        sb_mva: Base power in MVA.
        ib_ka: Base current in kA.

    Returns:
        Base impedance in ohms.
    """
    vbll_kv = float(vbll_kv)
    if ib_ka is None and sb_mva is not None:
        sb_mva = float(sb_mva)
        return (vbll_kv ** 2) / sb_mva
    elif sb_mva is None and ib_ka is not None:
        ib_ka = float(ib_ka)
        return vbll_kv / (ib_ka * 3 ** 0.5)
    else:
        raise ValueError("You must provide either ib_ka or sb_mva")


def s_base_mva(vbll_kv, ib_ka):
    """Returns the base power based on the base voltage and current.

    Args:
        vbll_kv: Base voltage in kV.
        ib_ka: Base current in kA.

    Returns:
        Base power in MVA.
    """
    vbll_kv = float(vbll_kv)
    ib_ka = float(ib_ka)
    return ib_ka * vbll_kv * 3 ** 0.5



from math import cos, pi, sin, sqrt
from numpy import matrix

builtin_complex = complex
from energy_tools.complex import complex


class Phasor(object):
    """New data type for the electrical phasor used in power engineering.

    A phasor is defined by an amplitude and a phase. The instance can be created either
    using those, or by providing a complex amplitude (in this case the phase is
    ignored). Several operations are supported, including: addition, substraction,
    multiplication, division, power, inversion and equality with either another phasor,
    a float or an integer.

    It also provides a nice representation in this form:

        120.000 @ 0.000°

    Attributes:
        amp: The unitless phasor amplitude.
        pha: The phasor's phase in degrees.
        real: The phasor's real part (interpreted as a complex number).
        imag: The phasor's imaginary part (interpreted as a complex number).
    """

    def __init__(self, amp, pha=0.0):
        """Initialize the phasor either using a real amplitude and a phase, or a
        complex amplitude.

        Args:
            amp: The unitless phasor amplitude.
            pha: The phasor's phase in degrees.
        """

        if amp.__class__ == builtin_complex:
            new_complex = complex(amp)
            self.amp = new_complex.module
            self.pha = new_complex.phase

        elif amp.__class__ == complex:
            self.amp = amp.module
            self.pha = amp.phase

        else:
            if amp.__class__ == int:
                amp = float(amp)

            if pha.__class__ == int:
                pha = float(pha)

            if amp >= 0:
                self.amp = amp
            elif amp < 0:
                self.amp = -amp
                pha += 180

            if pha < 0:
                self.pha = pha % -360 + 360
            else:
                self.pha = pha % 360

    @property
    def real(self):
        return self.amp * cos(self.pha * pi / 180)

    @property
    def imag(self):
        return self.amp * sin(self.pha * pi / 180)

    def __add__(self, other):
        if other.__class__ == int or other.__class__ == float:
            return self.__add__(Phasor(other, 0))

        elif other.__class__ == complex:
            return self.__add__(Phasor(other))

        elif other.__class__ == Phasor:
            x1 = self.amp * cos(self.pha * pi / 180)
            y1 = self.amp * sin(self.pha * pi / 180)
            x2 = other.amp * cos(other.pha * pi / 180)
            y2 = other.amp * sin(other.pha * pi / 180)
            a = sqrt(pow(x1 + x2, 2) + pow(y1 + y2, 2))
            p = complex(x1 + x2 + (y1 + y2) * 1j).phase
            return Phasor(a, p)

        else:
            return NotImplemented

    def __sub__(self, other):
        if other.__class__ == int or other.__class__ == float:
            return self.__sub__(Phasor(other, 0))

        elif other.__class__ == complex:
            return self.__sub__(Phasor(other))

        elif other.__class__ == Phasor:
            x1 = self.amp * cos(self.pha * pi / 180)
            y1 = self.amp * sin(self.pha * pi / 180)
            x2 = other.amp * cos(other.pha * pi / 180)
            y2 = other.amp * sin(other.pha * pi / 180)
            a = sqrt(pow(x1 - x2, 2) + pow(y1 - y2, 2))
            p = complex(x1 - x2 + (y1 - y2) * 1j).phase
            return Phasor(a, p)

        else:
            return NotImplemented

    def __mul__(self, other):
        if other.__class__ == int or other.__class__ == float:
            return self.__mul__(Phasor(other, 0))

        elif other.__class__ == complex:
            return self.__mul__(Phasor(other))

        elif other.__class__ == Phasor:
            return Phasor(self.amp * other.amp, self.pha + other.pha)

        else:
            return NotImplemented

    def __truediv__(self, other):
        if other.__class__ == int or other.__class__ == float:
            return self.__truediv__(Phasor(other, 0))

        elif other.__class__ == complex:
            return self.__truediv__(Phasor(other))

        elif other.__class__ == Phasor:
            return Phasor(self.amp / other.amp, self.pha - other.pha)

        else:
            return NotImplemented

    def __radd__(self, other):
        if other.__class__ == int or other.__class__ == float:
            return self.__add__(Phasor(other, 0))

        elif other.__class__ == complex:
            return self.__add__(Phasor(other))

        else:
            return NotImplemented

    def __rsub__(self, other):
        if other.__class__ == int or other.__class__ == float:
            x1 = self.amp * cos((self.pha + 180) * pi / 180)
            y1 = self.amp * sin((self.pha + 180) * pi / 180)
            x2 = other
            y2 = 0
            a = sqrt(pow(x1 + x2, 2) + pow(y1 + y2, 2))
            p = complex(x1 + x2 + (y1 + y2) * 1j).phase
            return Phasor(a, p)

        else:
            return NotImplemented

    def __rmul__(self, other):
        if other.__class__ == int or other.__class__ == float:
            return self.__mul__(Phasor(other, 0))

        elif other.__class__ == complex:
            return self.__mul__(Phasor(other))

        else:
            return NotImplemented

    def __rtruediv__(self, other):
        if other.__class__ == int or other.__class__ == float:
            a = other / self.amp
            if self.pha >= 180:
                p = self.pha - 180
            else:
                p = self.pha + 180
            return Phasor(a, p)

        else:
            return NotImplemented

    def __pow__(self, other, modulo=None):
        if modulo is not None:
            return NotImplemented

        elif other.__class__ == int:
            r = 1
            for i in range(other):
                r = self * r
            return r

        else:
            return NotImplemented

    def __abs__(self):
        return self.amp

    def __neg__(self):
        return self.__rsub__(0)

    def __pos__(self):
        return self

    def __invert__(self):
        return self.__neg__()

    def __int__(self):
        return int(self.amp)

    def __float__(self):
        return float(self.amp)

    def __complex__(self):
        return complex(self.real, self.imag)

    def __round__(self, n=0):
        return Phasor(round(self.amp, n), round(self.pha, n))

    def __repr__(self):
        s = str(round(self.amp, 3)) + " @ "
        s += str(round(self.pha, 3)) + "°"
        return s

    def __eq__(self, other):
        if other.__class__ == int or other.__class__ == float:
            other = Phasor(other, 0)
        elif other.__class__ != Phasor:
            return False
        elif round(self.amp, 2) != round(other.amp, 2):
            return False
        elif round(self.pha, 1) != round(other.pha, 1):
            return False
        else:
            return True

    def __ne__(self, other):
        return not self.__eq__(other)


def sequences(v_a, v_b, v_c):
    """Returns phase A's sequence voltages from phases A, B and C's voltages.

    Args:
        v_a: Phase A voltage.
        v_b: Phase B voltage.
        v_c: Phase C voltage.

    Returns:
        v_a0, v_a1, v_a2: Zero, positive and negative sequence voltages for phase A.
    """
    a = Phasor(1, 120)
    At = matrix(
        [[1 / 3, 1 / 3, 1 / 3], [1 / 3, a / 3, a ** 2 / 3], [1 / 3, a ** 2 / 3, a / 3]]
    )
    Vabc = matrix([[Va], [Vb], [Vc]])
    Va012 = At * Vabc
    return Va012[0].item(), Va012[1].item(), Va012[2].item()


def phasors(v_a0, v_a1, v_a2):
    """Returns phase A's sequence voltages from phases A, B and C's voltages.

    Args:
        v_a0: Zero sequence voltages for phase A.
        v_a1: Positive sequence voltages for phase A.
        v_a2: Negative sequence voltages for phase A.

    Returns:
        v_a, v_b, v_c: Phases A, B and C's voltages.
    """
    a = Phasor(1, 120)
    A = matrix([[1, 1, 1], [1, a ** 2, a], [1, a, a ** 2]])
    Va012 = matrix([[v_a0], [v_a1], [v_a2]])
    Vabc = A * Va012
    return Vabc[0].item(), Vabc[1].item(), Vabc[2].item()



from math import sqrt

builtin_complex = complex
from energy_tools.complex import complex


def test_init():
    assert complex(1, 2) == builtin_complex(1, 2)


def test_phase_simple():
    assert complex(1, 1).phase == 45


def test_phase_quadran_1():
    assert complex(1 + 1j).phase == 45.0


def test_phase_quadran_2():
    assert complex(-1 + 1j).phase == 135.0


def test_phase_quadran_3():
    assert complex(-1 - 1j).phase == 225.0


def test_phase_quadran_4():
    assert complex(1 - 1j).phase == 315.0


def test_phase_0_degres():
    assert complex(1 + 0j).phase == 0.0


def test_phase_90_degres():
    assert complex(0 + 1j).phase == 90.0


def test_phase_180_degres():
    assert complex(-1 + 0j).phase == 180.0


def test_phase_270_degres():
    assert complex(0 - 1j).phase == 270.0


def test_phase_entier_positif():
    assert complex(2).phase == 0.0


def test_phase_decimal_negatif():
    assert complex(-2.12).phase == 180.0


def test_module():
    assert complex(1, 1).module == sqrt(2)



def utilisation_factor(hours_list, p_list, p):
    """Returns the utilisation factor based on the active power during a year.

    Args:
        hours_list: List of duration in hours for each active power in p_list.
        p_list: List of active power at each hour.
        p: Nominal active power (single value).

    Returns:
        Utilisation factor in per unit.
    """
    if len(hours_list) != len(p_list):
        raise ValueError("hours_list and p_mw_list must have the same length.")

    p_pu = []
    for i in range(len(hours_list)):
        p_pu.append(p_list[i] / p)

    e_pu = []
    for i in range(len(hours_list)):
        e_pu.append(hours_list[i] * p_pu[i])

    return sum(e_pu) / (24 * 365.25)


def loss_factor(hours_list, p_list, p):
    """Returns the loss factor based on the active power during a year.

    Args:
        hours_list: List of duration in hours for each active power in p_list.
        p_list: List of active power at each hour.
        p: Nominal active power (single value).

    Returns:
        Loss factor in per unit.
    """
    p_pu = []
    for i in range(len(hours_list)):
        p_pu.append(p_list[i] / p)

    e2_pu = []
    for i in range(len(hours_list)):
        e2_pu.append(hours_list[i] * pow(p_pu[i], 2))

    return sum(e2_pu) / (24 * 365.25)
